package PureMessage::MIME::Encoding;

use strict;
use base 'Exporter';
our @EXPORT = qw(make_encoder make_decoder);

sub make_encoder {
    my $te = shift;
    $te =~ s/-/_/g;
    my $f = "make_" . lc($te) . "_encoder";
    return eval { no strict 'refs'; &$f };
}

sub make_decoder {
    my $te = shift;
    $te =~ s/-/_/g;
    my $f = "make_" . lc($te) . "_decoder";
    return eval { no strict 'refs'; &$f };
}

sub make_base64_decoder {
    require MIME::Base64;
    # Need a buffer because we need to make sure we decode
    # data in multiples of 4
    my $buf = "";
    sub {
	if (@_) {
	    $buf .= shift;
	}
	else {
	    # flush (add appropriate padding)
	    $buf .= "=" x (4 - (length($buf) % 4)) if length $buf;
	}
	$buf =~ tr[A-Za-z0-9+/=][]cd;
	my $mod = length($buf) % 4;
	local $^W;
	if ($mod) {
	    return MIME::Base64::decode(substr($buf, 0, -$mod, ""));
	}
	else {
	    my $tmp = MIME::Base64::decode($buf);
	    $buf = "";
	    return $tmp;
	}
    };
}

sub make_quoted_printable_decoder {
    require MIME::QuotedPrint;
    # Need a buffer here because we must ensure that we don't
    # split an encoded =XX sequence in two, and be immune to
    # extra space after trailing "=".  Always decoding complete
    # lines ensures that.
    my $buf = "";
    sub {
	return MIME::QuotedPrint::decode($buf) unless @_;  # flush
	$buf .= shift;
	$buf =~ s/\r\n/\n/g;  # workaround for bug(!) in MIME::QuotedPrint
	if ($buf =~ /[^\n]+\z/) {
	    return MIME::QuotedPrint::decode(substr($buf, 0, $-[0], ""));
	}
	elsif (length($buf)) {
	    # the whole buffer ends with "\n" so we can safely
	    # decode it all.
	    my $tmp = MIME::QuotedPrint::decode($buf);
	    $buf = "";
	    return $tmp;
	}
	else {
	    return "";
	}
    }
}

sub make_7bit_decoder {
    # a decoder that turns CRLF into LF, but leaves lone CR asis
    my $buf = "";  # empty or CR
    sub {
	return $buf unless @_;
	my $chunk = shift;
	if (length $buf) {
	    return "" unless length $chunk;
	    substr($chunk, 0, 0) = $buf unless $chunk =~ /^\n/;
	}
	$chunk =~ s/\r\n/\n/g;
	$buf = ($chunk =~ /\r\z/) ? chop($chunk) : "";
	return $chunk;
    }
}

*make_8bit_decoder = \&make_7bit_decoder;

sub identity {
    return "" unless @_;
    return $_[0];
}

sub make_binary_decoder {
    return \&identity;
}

*make_binary_encoder = \&make_binary_decoder;

sub make_base64_encoder {
    require MIME::Base64;
    # Need to encode the data in chunks with length that is a
    # multiple of 57 to get the end padding corrent and nice
    # line-up of the encoded lines.
    my $buf = "";
    sub {
	if (@_) {
	    $buf .= shift;
	    return MIME::Base64::encode(substr($buf, 0, length($buf) - (length($buf) % 57), ""), "\015\012");
	}
	else {
	    return MIME::Base64::encode($buf, "\015\012");
	}
    };
}

sub make_quoted_printable_encoder {
    require MIME::QuotedPrint;
    # Should encode whole lines in order to not introduce artifacs
    my $buf = "";
    sub {
	unless (@_) {
	    my $res = MIME::QuotedPrint::encode($buf);
	    $res =~ s/\n/\015\012/g;
	    return $res;
	}

	$buf .= shift;
	$buf =~ s/\r\n/\n/g;  # workaround for bug(!) in MIME::QuotedPrint
	if ($buf =~ /[^\n]+\z/) {
	    my $res = MIME::QuotedPrint::encode(substr($buf, 0, $-[0], ""));
	    $res =~ s/\n/\015\012/g;
	    return $res;
	}
	elsif (length($buf)) {
	    # the whole buffer ends with "\n" so we can safely
	    # decode it all.
	    my $res = MIME::QuotedPrint::encode($buf);
	    $res =~ s/\n/\015\012/g;
	    $buf = "";
	    return $res;
	}
	else {
	    return "";
	}
    };
}

sub make_8bit_encoder {
    # an encoder that turns LF into CRLF
    my $buf = "";
    sub {
	return $buf unless @_;
	my $chunk = shift;
	if (length $buf) {
	    return "" unless length $chunk;
	    substr($chunk, 0, 0) = $buf;
	}

	# Translate LF => CRLF, but not if the LF is
	# already preceeded by a CR (borrowed from Net::Cmd)
	$chunk =~ s/\G()\n|([^\r\n])\n/$+\015\012/sgo;

	$buf = ($chunk =~ /\r\z/) ? chop($chunk) : "";
	return $chunk;
    };
}

sub make_7bit_encoder {
    my $enc = make_8bit_encoder();
    sub {
	my $res = &$enc;
	$res =~ s/[^\0-\x7f]/?/g;
	$res;
    };
}

1;
