package PureMessage::MIME::Exploder;

use strict;

use PureMessage::MIME::Parser;

use base 'Exporter';
our @EXPORT_OK = qw(explode);

my @EXPL;

sub explode {
    my ($file, $dir, %opts) = @_;
    my $N = 0;
    my $F = sub { "$dir/$N" };

    # NOTE: explode() keeps a maximum of 2 open fds in @F at once: one for the
    # top-level message, and one for the current attachment or part.
    # Attachments are pushed onto @explode if they need to be exploded
    # further. In other words, we explode breadth first because it consumes
    # fewer filehandles (although it uses more disk space :)
    my (@explode, @F, @visit) = $file;
    my $p = PureMessage::MIME::Parser->new;
    $p->decode_te(1);
    $p->handler(part_start => sub {
	my (undef, undef, $ct) = @_;
	$N++;
	open(my $TMP, '>', &$F) or die "can't create ".&$F.": $!";
	binmode $TMP;
	push @F, $TMP;
	if ($ct =~ m#^(?:message|multipart)/#) {
	    push @explode, &$F;
	}
	push @visit, &$F;
    });
    $p->handler(data => sub {
	my (undef, $data) = @_;
	print { $F[-1] } $data or die "can't write to ".&$F.": $!";
    });
    $p->handler(part_end => sub {
	close $F[-1] or die "can't close ".&$F.": $!";
	pop @F;
    });

    my @files;
    while (@explode) {
	$p->parse_file(shift @explode);

	# Check for other encoding techniques on the extracted files:
	for my $EXPL (@EXPL) {
	    &$EXPL(\@visit, \@explode, \$N, $F);
	}

	# If they passed us (visit => sub {}), then visit the files in %visit
	# now, then unlink them. This means we visit files as soon as possible
	# and remove them immediately. If they didn't pass 'visit', just leave
	# everything until we return.
	if (my $cb = $opts{visit}) {
	    for my $f (@visit) {
		$cb->($f);
		unlink $f unless grep { $f eq $_ } @explode;
	    }
	}
	else { push @files, @visit }

	# Clear out the %visit hash for next time round.
	@visit = ();
    }
    @files;
}

@EXPL = (
    # Check for and explode UU-encoded files
    sub {
	my ($files, $explode, $N, $F) = @_;
	for my $f (@$files) {
	    my @F;
	    open(my $TMP, '<', $f) or die "can't read $f: $!";
	    binmode($TMP);
	    while (<$TMP>) {
		if (/^begin \d+ [^\n\r]+/) {
		    $$N++;
		    open(my $UU, '>', &$F) or die "can't write ".&$F.": $!";
		    binmode $UU;
		    push @F, $UU;
		    push @$explode, &$F;
		}
		elsif (/^end/ and @F) {
		    close($F[-1]) or die "can't close ".&$F.": $!";
		    pop @F;
		}
		elsif (@F) {
		    print { $F[-1] } unpack("u*", $_);
		}
	    }
	}
    },

    # TODO - tar, gzip, bzip2, etc.
);

1;

=head1 NAME

PureMessage::MIME::Exploder - explode MIME parts into a directory

=head1 SYNOPSIS

  use PureMessage::MIME::Exploder 'explode';

  mkdir("foo.d", 0755) || die;

  # Visit each file as it's created
  explode("foo.mail", "foo.d", visit => sub {
      my $f = shift;
      print "Visiting file $f\n";
  });

  # Or blow them all up and traverse yourself:
  @files = explode("foo.mail", "foo.d");

=head1 DESCRIPTION

PureMessage::MIME::Exploder recursively descends into multipart/* and
message/rfc822 MIME attachments. It uses PureMessage::MIME::Parser internally.
In addition to extracting MIME, it also extracts UU-encoded files (and
recurses into those files, too).

=head2 Using PureMessage::MIME::Exploder

The single entry point is explode():

   explode($file, $directory, %opts);

explode()'s arguments are defined below.

=over 4

=item $file

The file containing MIME that you want to explode.

=item $directory

A directory for explode() to use. It must exist.

=item %opts

The following options are recognized:

=over 4

=item visit

A CODE ref. If you supply C<visit>, your callback will be invoked as each part
is extracted, then the files will be unlink()ed. If not, explode() returns the
list of files it created (and doesn't delete them).

=back

=back

=head1 BUGS

There should be support for descending into the following content:

=over 4

=item zip

=item gzip

=item bzip2

=item tar

=item binhex

=back

The code is written so that it should be easy to drop in support for these
types when we need to.

=head1 AUTHOR

ActiveState Corporation

=head1 COPYRIGHT

Copyright (C) 2003, ActiveState Corporation.
All Rights Reserved.
