package PureMessage::MIME::MsgEditor;

use strict;
use PureMessage::MIME::Parser ();
use PureMessage::MIME::Encoding qw(make_encoder);
use Carp qw(carp croak);

#use Data::Dump qw(dump);

sub HEADER_CLASS () { "PureMessage::MIME::MsgEditor::Header" }
sub CHANGE_CLASS () { "PureMessage::MIME::MsgEditor::Change" }

# want a level of indirection in the pmx method names
# so that we can override them.  The PureMessage::Milter::Scanner needs
# this because it has already overridden the standard names
# in a non-standard way :-(
our($ADDHEADER, $CHGHEADER, $REPLACEBODY) = ("addheader", "chgheader", "replacebody");

sub new {
    my $class = shift;
    return bless { changes => [] }, $class;
}

sub clone {
    my $self = shift;
    my $other = bless {}, ref($self);
    for my $k (keys %$self) {
	if ($k eq "changes") {
	    $other->{$k} = [@{$self->{$k}}];
	}
	else {
	    $other->{$k} = $self->{$k}
	}
    }
    #dump($self, $other);
    return $other;
}

sub change {
    my $self = shift;
    push(@{$self->{changes}}, CHANGE_CLASS->new(@_));
    delete $self->{conflicts_checked};
    return;
}

sub find_conflicts {
    my $self = shift;

    my @ignored;
    my @accepted;
    my @changes = @{$self->{changes}};
    @changes = sort { length($a->{part_id}) <=> length($b->{part_id}) } @changes;

    while (@changes) {
	my $c1 = shift @changes;
	my $id1 = $c1->{part_id};
	#warn "Check $c1->{id} (part=$id1)\n";

	my $dead_change;
	my @dead_children;

	for my $c2 (@changes) {
	    my $id2 = $c2->{part_id};
	    my $conflict;
	    if ($id1 eq $id2) {
		#warn "affects same part as $c2->{id}";
		if ($c1->is_body_change && $c2->is_body_change) {
		    #warn "both affect the body";
		    $conflict++;
		}

		# if the changes touch the same header then it
		# is an conflict
		my %tmp = map {$_ => 1} $c1->changed_headers;
		if (%tmp) {
		    for my $h ($c2->changed_headers) {
			$conflict++ if $tmp{$h};
		    }
		}
	    }
	    elsif ("$id1." eq substr($id2, 0, length($id1)+1)) {
		#warn "affects subpart $c2->{id}";
		if ($c1->is_body_change) {
		    #warn "both affect the body of $c1->{id}";
		    $conflict++;
		}
	    }
	    else {
		#warn "not related";
	    }

	    if ($conflict) {
		# figure out which change has priority
		my $i1 = $c1->{importance} || 0;
		my $i2 = $c2->{importance} || 0;
		if ($i1 < $i2 || ($i1 == $i2 && $c1->{id} > $c2->{id})) {
		    $dead_change++;
		}
		else {
		    # throw out c2
		    push(@dead_children, $c2);
		}
	    }
	}

	if ($dead_change) {
	    push(@ignored, $c1);
	}
	else {
	    #warn "ACCEPT $c1->{id}\n";
	    push(@accepted, $c1);
	    if (@dead_children) {
		for my $c (@dead_children) {
		    $c->{killed_by} = $c1;
		    push(@ignored, $c);
		    @changes = grep { $_ != $c } @changes;
		}
	    }
	}
    }

    if (@ignored) {
	#warn "IGNORED: ", join(", ", map $_->{id}, @ignored), "\n";
    }

    @accepted = sort {$a->{id} <=> $b->{id}} @accepted;

    $self->{changes} = \@accepted;
    $self->{conflicts_checked}++;

    return @ignored;
}

sub do_changes {
    my($self, $msg_file, $pmx) = @_;

    $self->find_conflicts unless $self->{conflicts_checked};

    my @all_changes = @{$self->{changes}};
    return unless @all_changes;

    local $ADDHEADER = $self->{addheader} if $self->{addheader};
    local $CHGHEADER = $self->{chgheader} if $self->{chgheader};
    local $REPLACEBODY = $self->{replacebody} if $self->{replacebody};
    
    my $p = PureMessage::MIME::Parser->new;
    $p->{stack} = [];

    $p->handler(part_start => sub {
	my($self, $hdr, $ct, $ct_param) = @_;

	$self->{part_closed} = 0;

	my $stack = $self->{stack};
	my $pframe;
	$pframe = $stack->[-1] if @$stack;
	my %frame = (ct => $ct, ct_param => $ct_param);
	push(@$stack, \%frame);

	#dump($self);

	my $part_id = $self->part_id;
	my @part_changes = grep {$_->{part_id} eq $part_id} @all_changes;

	# delete
	for my $c (@part_changes) {
	    next unless $c->{delete};
	    $self->{suppress_data}++;
	    $self->decode_multipart(0);
	    return;
	}

	# init boundary
	if ($ct =~ m,^multipart/, && exists $ct_param->{boundary}) {
	    $frame{boundary} = $ct_param->{boundary};
	}

	# output boundary?
	if ($pframe && exists $pframe->{boundary}) {
	    # XXX should not prefix with "\r\n" if it is the first
	    # one at this level and no prolog data has been put out
	    $pmx->$REPLACEBODY("\r\n--$pframe->{boundary}\r\n");
	}

	unless (@part_changes) {
	    # no need for talking time to create an header object
	    # just to print it out.
	    if ($part_id ne "1") {
		$hdr =~ s/\G()\n|([^\r\n])\n/$+\015\012/sgo;  # LF --> CRLF
		$pmx->$REPLACEBODY($hdr);
	    }
	    return;
	}


	$hdr = HEADER_CLASS->new($hdr);
	my $orig_hdr;
	$orig_hdr = $hdr->dup if $part_id eq "1";

	# change headers
	{
	    my @chg_hdr;
	    for my $c (@part_changes) {
		my $ch = $c->{change_headers};
		next unless $ch;
		push(@chg_hdr, @$ch);
	    }

	    if (@chg_hdr) {
		# sort it with highest indexes first
		@chg_hdr = sort {($b->[2] || 1) <=> ($a->[2] || 1) } @chg_hdr;

		# do the update
		for my $ch (@chg_hdr) {
		    $hdr->replace(@$ch);
		}
	    }
	}

	# add headers
	for my $c (@part_changes) {
	    my $ah = $c->{add_headers};
	    next unless $ah;
	    for my $h (@$ah) {
		$hdr->add(@$h);
	    }
	}

	# deal with attachments
	{
	    my @attach;
	    for my $c (@part_changes) {
		my $ae = $c->{add_entities};
		next unless $ae;
		push(@attach, @$ae);
	    }

	    if (@attach) {
		$frame{attach} = \@attach;

		if ($ct !~ m/^multipart/) {
		    my $new_hdr = $hdr->dup;
		    for my $t ($new_hdr->tags) {
			((lc($t) =~ /^content-/) ? $hdr : $new_hdr)->delete($t);
		    }
		    my $b = $frame{boundary2} = _new_boundary();
		    $hdr->replace("Content-Type", qq(multipart/mixed; boundary="$b"));
		    
		    # output header
		    $hdr->pmx_out($pmx, $orig_hdr);
		    $pmx->$REPLACEBODY("--$b\r\n");

		    $hdr = $new_hdr;
		    $orig_hdr = undef;
		}
	    }

	}

	# change body
	my $change_body;
	for my $c (@part_changes) {
	    last if ($change_body = $c->{change_body});
	}

	my $body_enc;
	$body_enc = _body_enc($hdr, $change_body->[1]) if $change_body;

	# output header
	$hdr->pmx_out($pmx, $orig_hdr);

	if ($change_body) {
	    my $flag = $change_body->[2];
	    $flag .= "r" if !defined($flag) || $flag !~ /[apr]/;
	    my $is_replace = $flag =~ /r/;

	    if ($flag =~ /[pr]/) {
		_body_repl($pmx, $body_enc, $is_replace, $change_body->[0]);
	    }
	    else {
		# must be append then, so we can't run _body_repl
		# until the part_end callback.
		$self->{delayed_body_repl} = $change_body;
	    }

	    if ($is_replace) {
		$self->{suppress_data}++;
		$self->decode_multipart(0);
	    }
	    else {
		$self->{body_encoder} = $body_enc;
		$self->decode_multipart(0);
		$self->decode_te(1) if $body_enc;
	    }
	}
	elsif ($part_id eq "1" &&
	       !$frame{attach} &&
	       !grep {$_->{part_id} ne "1"} @all_changes)
	{
	    # we know that there is no changes that affect the body, so there is
	    # no need to try to rewrite the body either; bail our right away
	    die "DONE";
	}

	return;

    });

    $p->handler(part_end => sub {
	my($self) = @_;

	my $stack = $self->{stack};
	my $frame = pop(@$stack);

	#dump($frame);

	if (my $change_body = delete $self->{delayed_body_repl}) {
	    _body_repl($pmx, delete $self->{body_encoder}, 1, $change_body->[0]);
	}
	elsif (my $enc = delete $self->{body_encoder}) {
	    # flush it
	    my $buf = &$enc();
	    $pmx->$REPLACEBODY($buf) if length $buf;
	}

	# output attachments
	if (my $attach = $frame->{attach}) {
	    my $b = exists $frame->{boundary2} ? $frame->{boundary2}
	                                       : $frame->{boundary};

	    for my $a (@$attach) {
		$pmx->$REPLACEBODY("--$b\r\n");
		my $hdr = HEADER_CLASS->new($a->[0]);
		my $enc = _body_enc($hdr, $a->[2]);
		$pmx->$REPLACEBODY($hdr->as_string);
		_body_repl($pmx, $enc, 1, $a->[1]);
	    }

	    if (exists $frame->{boundary2}) {
		$pmx->$REPLACEBODY("\r\n--$frame->{boundary2}--\r\n");
	    }
	}

	if ($self->{part_closed}) {
	    die unless exists $frame->{boundary};
	    $pmx->$REPLACEBODY("\r\n--$frame->{boundary}--\r\n");
	}
	$self->{part_closed}++;

	# reset
	delete $self->{suppress_data};
	$self->decode_multipart(1);
	$self->decode_te(0);
    });

    $p->handler(data => sub {
	return if $_[0]{suppress_data};
	if (my $enc = $_[0]{body_encoder}) {
	    my $buf = &$enc($_[1]);
	    $pmx->$REPLACEBODY($buf) if length $buf;
	}
	else {
	    $pmx->$REPLACEBODY($_[1]);
	}
    });

    eval {
        if ( ref($msg_file) eq "ARRAY" ) {
            foreach (@{$msg_file}) { 
                $p->parse($_);
            }
        }
        else {
            $p->parse_file($msg_file);
        }
    };
    Carp::croak($@) if $@ && $@ !~ /^DONE /;
    
    return;
}

my $b_count;
sub _new_boundary {
    $b_count++;
    return "==" . join("-", time, $$, $b_count) . "==";
}

sub _body_enc {
    my($hdr, $te) = @_;
    my $enc;
    if ($te) {
	$enc = make_encoder($te);
	unless ($enc) {
	    $te = "base64";
	    $enc = make_encoder($te);  # always works
	}
	$hdr->replace("Content-Transfer-Encoding", $te);
    }
    return $enc;
}

sub _body_repl {
    my $pmx = shift;
    my $enc = shift;
    my $flush = shift;

    if (ref($_[0])) {
	# assume it is a file
	my $n;
	my $f = $_[0];
	my $buf;
        my $seek = tell($f);
	while ( ($n = read($f, $buf, 1024))) {
	    $buf = &$enc($buf) if $enc;
	    $pmx->$REPLACEBODY($buf) if length $buf;
	}
        seek($f, $seek, 0);
	if ($enc && $flush) {
	    $buf = &$enc();
	    $pmx->$REPLACEBODY($buf) if length $buf;
	}
    }
    elsif ($enc) {
	if ($flush) {
	    $pmx->$REPLACEBODY(join("", &$enc($_[0]), &$enc()));
	}
	else {
	    $pmx->$REPLACEBODY(&$enc($_[0]));
	}
    }
    else {
	$pmx->$REPLACEBODY($_[0]);
    }
}

sub _cmp_change_by_part_id {
    local $a = $a->{part_id};
    local $b = $b->{part_id};
    return &_cmp_part_id;
}

sub _cmp_part_id {
    #warn "CMP $a $b\n";
    return 0 if $a eq $b;  # quick exit for common case
    my $aid = $a;
    my $bid = $b;
    while (length $aid) {
	return 1 unless length $bid;
	$aid =~ s/^(\d+)\.?// || die;
	my $aa = $1;
	$bid =~ s/^(\d+)\.?// || die;
	my $bb = $1;
	next if $aa == $bb;
	return $aa <=> $bb;
    }
    return -1 if length $bid;
    return 0;
}

package PureMessage::MIME::MsgEditor::Change;

my $change_id;

sub new {
    my $class = shift;
    my %self;
    while (@_) {
	my $k = shift;
	my $v = shift;
	if ($k eq "add_header" || $k eq "change_header") {
	    unless (ref $v) {
		$v = [split(/\s*:\s?/, $v)];
		$v->[1] = undef if defined($v->[1]) && $v->[1]  eq "";
	    }
	    push(@{$self{$k."s"}}, $v);
	}
	elsif ($k eq "add_entity") {
	    unless (ref $v) {
		$v = [split(/\r?\n\r?\n/, $v)];
	    }
	    push(@{$self{add_entities}}, $v);
	}
	elsif ($k eq "add_headers" || $k eq "change_headers" || $k eq "add_entities") {
	    push(@{$self{$k}}, @$v);
	}
	else {
	    Carp::carp("Duplicate value for $k") if exists $self{$k};
	    $self{$k} = $v;
	}
    }
    $self{part_id} ||= 1;
    $self{id} = ++$change_id;
    bless \%self, $class;
}

sub id {
    shift->{id};
}

sub part_id {
    shift->{part_id};
}

sub is_body_change {
    my $self = shift;
    return !!($self->{change_body} || $self->{delete});
}

sub changed_headers {
    my $self = shift;
    my $ch = $self->{change_headers};
    return () unless $ch;
    return map {lc(join(":", $_->[0], ($_->[2] || 1)))} @$ch;
}

sub is_header_change {
    my $self = shift;
    return !!($self->{add_headers} || $self->{change_headers} || $self->{delete});
}


package PureMessage::MIME::MsgEditor::Header;

use base 'Mail::Header';

sub new {
    my $class = shift;
    $class->SUPER::new(@_ ? [split(/^/, $_[0])] : ());
}

sub as_string {
    my $self = shift;
    my $str = $self->SUPER::as_string;
    $str =~ s/\r//g;
    $str =~ s/\n/\r\n/g;
    $str .= "\r\n";
    $str;
}

sub replace {
    my($self, $tag, $line, $index) = @_;
    $index ||= 1;
    $index--;
    $self->{pmx_mod}++;
    if (defined $line) {
	$self->SUPER::replace($tag, $line, $index);
    }
    else {
	$self->SUPER::delete($tag, $index);
    }
}

sub delete {
    my $self = shift;
    my $tag = shift;
    $self->{pmx_mod}++;
    if (@_) {
	$self->SUPER::delete($tag, ($_[0] || 1)-1);
    }
    else {
	$self->SUPER::delete($tag);
    }
}

sub add {
    my $self = shift;
    $self->{pmx_mod}++;
    $self->SUPER::add(@_);
}

sub get {
    my $self = shift;
    my $tag = shift;
    if (@_) {
	$self->SUPER::get($tag, $_[0]-1);
    }
    else {
	$self->SUPER::get($tag);
    }
}

sub pmx_out {
    my($self, $pmx, $orig) = @_;
    return $pmx->$REPLACEBODY($self->as_string) unless $orig;

    return unless $self->{pmx_mod};

    # all this just to figure out the diff between $self and $orig
    my %cur_tags  = map {$_ => 1 } $self->tags;
    my %orig_tags = map {$_ => 1 } $orig->tags;

    for my $tag (sort keys %cur_tags) {
	if ($orig_tags{$tag}) {
	    # any change?
	    my @cur_vals = $self->get($tag);
	    my @orig_vals = $orig->get($tag);
	    my @extra_vals;

	    if (@orig_vals < @cur_vals) {
		push(@extra_vals, splice(@cur_vals, @orig_vals));
	    }
	    else {
		push(@cur_vals, ((undef) x (@orig_vals - @cur_vals)))
	    }

	    my $i = @orig_vals;
	    for my $ov (reverse @orig_vals) {
		my $v = pop @cur_vals;
		if (!defined($v) || $v ne $ov) {
		    chomp($v);
		    $pmx->$CHGHEADER($tag, $i, $v);
		}
		$i--;
	    }

	    for my $v (@extra_vals) {
		chomp($v);
		$pmx->$ADDHEADER($tag, $v);
	    }
	}
	else {
	    # added header
	    for my $v ($self->get($tag)) {
		# XXX we loose original casing of $tag here
		chomp($v);
		$pmx->$ADDHEADER($tag, $v);
	    }
	}
    }

    for my $tag (keys %orig_tags) {
	unless ($cur_tags{$tag}) {
	    # this one was deleted
	    my $n = $orig->count($tag);
	    for my $i (reverse 1 .. $n) {
		$pmx->$CHGHEADER($tag, $i, undef);
	    }
	}
    }

    return;
}

1;

=head1 NAME

PureMessage::MIME::MsgEditor - Message update manager

=head1 SYNOPSIS

  use PureMessage::MIME::MsgEditor;
  $e = PureMessage::MIME::MsgEditor->new;

  # register changes
  $e->change(change_header => ["Subject", "Hello"]);
  $e->change(add_entity => $body, part_id => "1.1");

  # optional
  @ignored_changes = $e->find_conflicts;

  # propegate changes as milter callbacks
  $e->do_changes($msg_file, $pmx);

=head1 DESCRIPTION

Instances of the C<PureMessage::MIME::MsgEditor> class keeps a set of change
requests for an e-mail message.  It is able to sort out conflicting
change requests and also invoke the appropriate PerlMx methods to make
sendmail modify the message in the requested way.

The following methods are provided:

=over

=item $e = PureMessage::MIME::MsgEditor->new

This creates a new C<PureMessage::MIME::MsgEditor> object.  No parameters
accepted.

=item $e->change( $key => $value,... )

This registers a new change request with the editor.  This summaries
the key/value paris accepted.  The C<add_*> and C<change_header> fields can
be repeated.

  add_header    => [$field, $value, $index],
  change_header => [$field, $value, $index],
  add_entity    => [$hdr, $body, $te],
  change_body   => [$body, $te, $flag],
  delete        => 1,
  part_id       => $id,
  importance    => $number,

The $hdr is a multi-line string containing RFC822 headers.  The $body
is either a literal string or a file handle.  The $te is a
C<Content-Transfer-Encoding> name (like "7bit", "base64" or
"quoted-printable") or C<undef>.  If it is C<undef>, then it is
assumed that $hdr contains the appropriate encoding declaration and
that bytes in $body is already properly encoded.  The $flag argument
can currently contain one of the letters "a", "p", "r"; meaning
"append", "prepend" or "replace".  The default flag is "r".

=item $e->find_conclicts

This method looks through the registered changes and find all changes
that are in conflict.  Conflicts arise if two changes try to modify
the same header or body part.  Adding headers and entities never
creates conflict by itself, but changes are treated as units and if
some of the edits it makes is in conflict the whole change is.

Conflicing changes are sorted out according to priority.  The
C<importance> field of change (default 0) is the first criteria,
second is registration order.  The loosing changes are removed from
the registered change list and returned.

=item $e->do_changes( $msg_file, $pmx )

This method will combine all the requested changes into a series of
method calls on the $pmx object.  The methods called will be
addheader(), chgheader() and replacebody() as described in L<PerlMx>.

The $msg_file should be the name of a file containing the base message
that edits are based on.

If the find_conflicts() method has not been called, before this method
is invoked it will be called by do_changes() and any conflicting
changes will be lost.

=back

=head1 BUGS

None.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<PureMessage::MIME::Parser>

=cut
