package PureMessage::MIME::Parser;

# Prototype streaming MIME::Parser where parts will
# be moved to XS to make it faster than fast.

use strict;
our $VERSION = '0.02';

use Carp qw(carp croak);
my $FIELD_NAME_RE = qr/[\x21-\x39\x3b-\x7e]+[ \t]*:/;

use PureMessage::MIME::Encoding qw(make_decoder);

BEGIN {
    # Try to import
    eval {
	require PureMessage::_sys;
	require Fcntl;
	Fcntl->import qw(O_RDONLY);
    };
    warn $@ if $@ && exists $ENV{PMX_MILTERNAME};
}

sub new {
    my $class = shift;
    bless {
	pmxmp_buf => '',             # buffer until event can be generated
	pmxmp_boundary_stack => [],  # track nested multipart/*
	pmxmp_hdr_expected => 1,     # still have not seen the header
	pmxmp_decode_te => 0,        # undo transfer encoding
	pmxmp_decode_multipart => 1, # descend into nested multipart/*
    }, $class;
}


#
# feeders
#

sub parse {
    my $self = shift;
    my $bstack = $self->{pmxmp_boundary_stack};

    for ($self->{pmxmp_buf}) {       # alias buffer as $_ (avoids copy)
	$_ = join("", $_, @_);       # append next chunk(s) to buffer

      PARSE:
	while (length $_) {

	    if ($self->{pmxmp_hdr_expected}) {
		my $hdr = "";
		# If it starts out looking like a boundary we assume the
		# header is empty.  RFC 2822 actually seems to allow
		# fields with names like "--Foo: foo", so this hack is not
		# really correct.  The 'field-name' production allows any
		# chars except space and ':'.  To be perfectly correct we
		# would have to only cut the header where one of the complete
		# in-scope boundaries are seen.
		unless (/^--/) {
		    # Can we find a complete header (everything until
		    # the first blank line)
		    if (s/^((?:|.*?\r?\n)\r?\n)//s) {
			$hdr = $1;
			# If we find something inside the might be the start
			# of a new boundary then put it back
			if ($hdr =~ /\r?\n--/) {
			    substr($_, 0, 0) = substr($hdr, $-[0], length($hdr), "");
			}
		    }
		    else {
			# If we find something that might be the start of
			# a new boundary, then we cut it there.  Otherwise
			# we wait for more data to arrive.
			return unless /\r?\n--/;
			$hdr = substr($_, 0, $-[0], "");
		    }
		}
		
		_hdr($self, $hdr);

		last PARSE unless length $_;
	    }
	    if (@$bstack) {
		# If we are inside some multipart/* part, then we need to
		# look for boundaries in the data to know when the parts
		# end (and possibly a new begin).  RFC 2046 (sec 5.1.2)
		# says that we must deal with entities missing a
		# teminating boundary marker.  We must therefore look
		# for all outstanding boundary markers.
		my $depth;
		for my $b (reverse @$bstack) {
		    my $boundary = $b->[0];
		    $depth++;
		    if ($_ =~ $boundary) {
			# found a boundary marker, fire event for all
			# data found before the marker.
			my $is_end_boundary = $1;
			$self->data(substr($_, 0, $-[0], "")) if $-[0];
			my $btext = substr($_, 0, $+[0] - $-[0], "");
			
			# depending on which marker we found we might need
			# to close some parts
			$depth-- unless $is_end_boundary;

			for (1 .. $depth) {
			    $self->part_end;
			    pop(@$bstack);
			}

			# finish up and restart the parsing
			unless ($is_end_boundary) {
			    $self->part_end if $b->[1] > 0;
			    $b->[1]++;
			    $self->{pmxmp_hdr_expected}++;
			}
			$self->boundary($btext);
			next PARSE;
		    }
		}

		# no boundary found, try to get rid of all the data that
		# can't possibly be part of a new incomplete boundary marker.

                if ( /(\r?\n|^)--[^\n]*\z/ or /(\r|\r?\n-?|^-)\z/ ) {
		    $self->data(substr($_, 0, $-[0], "")) if $-[0];
		    last PARSE;
                }
	    }

	    # it is safe to flush the whole buffer as data
	    $self->data($_);
	    $_ = '';
	    last PARSE;
	}
    }
}

sub eof {
    my $self = shift;

    if ($self->{pmxmp_hdr_expected}) {
	# badly formed message, assume all accumulated data
	# represents the header.  The _hdr() function will put
	# data back if we are wrong.
	my $tmp = $self->{pmxmp_buf};
	$self->{pmxmp_buf} = "";
	_hdr($self, $tmp);
    }

    # flush outstanding data
    if (length($self->{pmxmp_buf})) {
	$self->data($self->{pmxmp_buf});
	$self->{pmxmp_buf} = '';
    }

    # outstanding parts must be closed
    while (@{$self->{pmxmp_boundary_stack}}) {
	$self->part_end() if $self->{pmxmp_boundary_stack}[-1][1];
	pop(@{$self->{pmxmp_boundary_stack}});
    }

    # close the top-level message as well
    $self->part_end();

    # reset parser so that it can start parsing again
    $self->{pmxmp_hdr_expected} = 1;

    return;
}

sub parse_file {
    my($self, $file) = @_;

    if (defined &PureMessage::_sys::open) {
	my $fd = PureMessage::_sys::open($file, O_RDONLY());
	croak("Can't open $file: $!") if $fd < 0;
	eval {
	    my $buf;
	    my $n;
	    while ( ($n = PureMessage::_sys::read($fd, $buf, 1024)) > 0) {
		$self->parse($buf);
	    }
	    warn "read from $file failed: $!" if $n < 0;
	};
	PureMessage::_sys::close($fd);
	die if $@;  # rethrow;
    }
    else {
	# fall back to the standard perl handlers
	open(my $f, "<", $file) || croak("Can't open $file: $!");
	binmode($f);
	my $buf;
	my $n;
	while ( ($n = read($f, $buf, 1024))) {
	    $self->parse($buf);
	}
	warn "read from $file failed: $!" unless defined $n;
    }

    $self->eof;
}


#
# status reporting methods
#

sub part_id {
    my $self = shift;
    join(".", "1", grep $_, map $_->[1], @{$self->{pmxmp_boundary_stack}});
}

#
# get/set attributes
#

sub decode_te {
    my $self = shift;
    my $old = $self->{pmxmp_decode_te};
    $self->{pmxmp_decode_te} = shift if @_;
    $old;
}

sub decode_multipart {
    my $self = shift;
    my $old = $self->{pmxmp_decode_multipart};
    $self->{pmxmp_decode_multipart} = shift if @_;
    $old;
}


#
# callbacks
#

sub handler {
    my $self = shift;
    my $event = shift;
    croak("Unknow event") unless $event =~ /^(part_(start|end)|data|boundary)$/;
    $event = "pmxmp_" . $event . "_h";
    my $old = $self->{$event};
    $self->{$event} = shift if @_;
    $old;
}

sub part_start {
    _cb("part_start", @_);
}

sub data {
    if (my $d = $_[0]{pmxmp_decoder}) {
	my $data = &$d($_[1]);
	return _cb("data", $_[0], $data) if length $data;
	return;
    }
    
    _cb("data", @_);
}

sub part_end {
    if (my $d = delete $_[0]{pmxmp_decoder}) {
	my $data = &$d(); # flush
	_cb("data", $_[0], $data) if length $data;
    }
    _cb("part_end", @_);
}

sub boundary {
    _cb("boundary", @_);
}

sub _cb {
    my $cb = shift;
    my $self = shift;

    if ($self->{trace}) {
	require Data::Dump;
        print STDERR $cb . " " . Data::Dump::dump(@_) . "\n";
    }

    $cb = $self->{"pmxmp_" . $cb . "_h"} || return;
    return ref $cb ? &$cb($self, @_) : $self->$cb(@_);
}

sub _hdr {
    my($self, $hdr) = @_;
    my $bstack = $self->{pmxmp_boundary_stack};

    # This sub is called with $hdr containing everything up to
    # the first /^--/m or /^$/m, or EOF.

    if ($hdr !~ /^(?:$FIELD_NAME_RE|From |\r?\n)/o) {
	# it starts out with somethings that does not look like
	# it can be a header at all; put it all back.
	substr($self->{pmxmp_buf}, 0, 0) = $hdr;
	$hdr = "";
    }
    elsif ($hdr =~ /\n(?![ \t]|$FIELD_NAME_RE|\r?\n)/) {
	# it contains text that can't possibly be the continuation
	# of a header; put that part back.
	substr($self->{pmxmp_buf}, 0, 0) = substr($hdr, $-[0]+1, length($hdr), "");
    }

    # Look at the content type and encoding to figure out what to do.
    my($ct, %ct_param) = _content_type($hdr);
    my $te = _content_transfer_encoding($hdr);

    # Fill in defaults
    unless (defined $ct) {
	$ct = @$bstack && $bstack->[-1][2] eq "multipart/digest" ?
	    "message/rfc822" : "text/plain";
    }
    $ct_param{charset} ||= "US-ASCII" if $ct =~ m,^text/,;
    if (defined $te) {
	# fix some common misspellings
	$te =~ s/-// if $te eq "7-bit" || $te eq "8-bit";
    }
    else {
	$te = "7bit";
    }

    # Fire off the event
    $self->part_start($hdr, $ct, \%ct_param, $te);

    # Prepare decoding of subparts or data
    $self->{pmxmp_hdr_expected} = 0;
    if ($self->{pmxmp_decode_multipart} && $ct =~ m,^multipart/,) {
	if (defined(my $b = $ct_param{boundary})) {
	    push(@$bstack, [qr/(?:\r?\n|^)--\Q$b\E(--)?[ \t]*\r?\n/,
			    0, $ct]);
	}
    }

    if ($self->{pmxmp_decode_te} && $te ne "binary") {
	# Install decoder
	if (my $decoder = make_decoder($te)) {
	    $self->{pmxmp_decoder} = $decoder;
	}
	else {
            # be silent
	    # carp("Don't know how to decode '$te'");
	}
    }
}


#
# helper functions
#

sub _content_type {
    my $hdr = shift;

    # Extract the value of the content-type header.  The RE is
    # complicated by the fact that we want to mach broken lines
    return unless $hdr =~ /^Content-Type[ \t]*:[ \t]*([^\n]*(?:\n[\t ][^\n]*)*)/mi;
    my $val = $1;

    # Extract the content type.
    return unless $val =~ s/^\s*([^\s;]+)//;  # XXX spaces in the type spec
    my $ct = lc($1);
    return $ct unless wantarray;

    # Extract parameters
    $val =~ s/\r//g;
    my @params;
    while ($val =~ s/^\s*;\s*([^\s=;]+)//) {
	push(@params, lc($1));  # attribute name

	# a quoted value
	if ($val =~ s/^\s*=\s*\"([^\"\\]*(?:\\.[^\"\\]*)*)\"//) {
	    my $v = $1;
	    $v =~ s/\\(.)/$1/g;
	    push(@params, $v);
	# some unquoted value
	} elsif ($val =~ s/^\s*=\s*([^;\s]*)//) {
	    my $v = $1;
	    $v =~ s/\s+$//;
	    push(@params, $v);
	# no value, a lone token
	} else {
	    push(@params, undef);
	}
    }

    return $ct, @params;
}

sub _content_transfer_encoding {
    my $hdr = shift;
    # Extract the value of the content-transfer-encoding header
    return unless $hdr =~ /^Content-Transfer-Encoding[ \t]*:[ \t]*([^\n]*(?:\n[\t ][^\n]*)*)/mi;
    my $encoding = $1;
    $encoding =~ s/\s+//g;
    return lc($encoding);
}

1;

__END__

=head1 NAME

PureMessage::MIME::Parser - Event based parser for MIME messages

=head1 SYNOPSIS

  use PureMessage::MIME::Parser;

  $p = PureMessage::MIME::Parser->new;
  $p->handler("part_start" => \&part_start);
  $p->handler("part_end"   => \&part_end);
  $p->handler("data"       => \&data_h);

  # Parse message chunk by chunk
  $p->parse($chunk1);
  $p->parse($chunk2);
  #...
  $p->eof; # signal end of document

  # Parse directly from file
  $p->parse_file("message");

=head1 DESCRIPTION

Instances of the C<PureMessage::MIME::Parser> class can be used to break up
and decode messages that follow the RFC822/MIME format used for
Internet e-mail and USENET News.  MIME is described in RFC 2045 and RFC 2046.
The RFC822 format is described in RFC 2822.

In contrast to the CPAN module C<MIME::Parser> this module has a much
simpler API (and implementation), it does not require stream objects
to parse from, it is totally event based, and there are never any
temporary files created behind the scenes.  The difference is in many
ways similar to the difference between SAX and DOM parsers in the XML
world.

A message to be parsed is given to the C<PureMessage::MIME::Parser> in
arbitrary chunks by calling the parse() method as many times as
needed.  The end of message is then signaled by calling eof().  The
parser object will as a result of the data it is feed invoke
registered callback handlers.  The callbacks are invoked as early as
possbile to minimize the need for internal buffering in the parser
object.

The parser object is a hash and can be used to store state during
parsing.  Hash keys prefixed with "pmxpm_" are reserved for the parser
itself.

The following methods are provided:

=over

=item $p = PureMessage::MIME::Parser->new;

The constructor method is used to create new instances of the
C<PureMessage::MIME::Parser> class.  Currently it does not take any
arguments.  It will be extended so that initial handlers and parser
attribute values could be set up here.

=item $p->handler( event => \&subroutine )

=item $p->handler( event => method_name )

=item $p->handler( event => undef )

=item $p->handler( event )

The event method is used to register event handlers with the parser.
There are currently 4 kinds of events that can be reported:

  part_start($p, $hdrs, $ct, \%ct_param, $te)
  part_end($p)
  data($p, $data)
  boundary($p, $bdata)

All event handlers receive a reference to the parser object ($p) as
their first argument.

The C<part_start> event is called with the all the headers fields as
the first additional argument.  This is a single string containing all
the headers.  Then the parsed out and normalized C<Content-Type> and
any extra parameters that went with it as well as the normlized
C<Content-Transfer-Encoding> value.  The top level message itself is also
reported as a C<part_start> event.

The C<part_end> event does not receive any extra arguments.  It is
called to signal the end of a part.  The C<part_start> and C<part_end>
pairs will always be called such that they nest properly.

The C<data> event is called to report body data inside a part.  The
extra argument is a chunk of data.  No guaranties are made about how
this data is divided up between chunks, but the $data argument will
never be empty.

The C<boundary> event is called to report any boundary markers during
processing.  It is only useful to applications that want to make sure
they filter a message with the minimal amount of changes, as this
allow the application to preserve trailing whitespace and notice if
ending boundaries happen to be missing.

The concatenation of all the $hdrs, $data and $bdata arguments
reported will give exactly the original message (given that decode_te
parser attribute is false so that $data is not decoded).

The return value of handler callbacks are ignored.

The return value of the handler method is the old callback value.  The
callback value can be set to C<undef>.  The effect is the same as
setting it to C<sub {}>, but it is a more efficient way to ignore
events.

=item $p->parse( $string )

Parse $string as the next chunk of the message.

=item $p->eof

When all the chunks of the message has been reported with the parse()
method, then the eof() needs to be invoked to flush any buffered data
and also invoke the final C<end_part> events.

=item $p->parse_file( $filename )

This is a convenience method that parse data from the file given as
argument.  Currently the argument must be a file name, but it will be
extended to also allow already opened file handlers in the future.

=item $p->part_id

This reports a short string that identifies this part in the the
multipart hierarchy.  The top level message is identified as "1".  The
first part in the top level message is "1.1" and the second is "1.2".
The first part inside the second part of the top level message is
"1.2.1".

=item $p->decode_te

=item $p->decode_te( $bool )

This boolean attribute will if TRUE instruct the parser to decode body
data before it is reported to the C<data> event.  This will undo any
"Quouted-Printable" or "Base64" transfer encoding that might have been
applied to this part.  If the transfer encoding is "7Bit" or "8bit",
then "\r\n" is translated to normalized text where lines are separated
by a lone "\n".

The default value is FALSE.  It means that the C<data> event receives
the raw chars that make up the original message.

=item $p->decode_multipart

=item $p->decode_multipart( $bool )

This boolean attribute will if TRUE instruct the parser to traverse
down into nested body parts.  The default is TRUE.

If it is FALSE, then the part body will be reported as C<data>
instead.

=back

=head1 BUGS

Should we allow content types with spaces in them "text / plain"?  The
RFC seems to allow that.

Is boundary="" really legal?

When parsing the Content-Type or Content-Transfer-Encoding fields we
should allow comments everywhere whitespace can go.  Header comments are
non-trivial to handle in Perl since they nest.

Should be a way to automatically decode nested "message/rfc822" parts.
Should there be a separate 'decode_xxx' attribute for it?

Should be support to descend into uuencoded stuff as well as tar and zip
files.

Should be support for pushing additional decoders in the C<part_start>
callback.  Could for instance be used to deal with automatic
decompression.

There is no way to get at the raw message data when C<decode_te> is enabled.

The top level C<part_start> can be called with an empty headers
argument.  This is not a valid RFC822 message.

There should be a pull-style interface to this parser as well.

It should be faster.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<pmx-intro>, L<PerlMx>, L<PerlMx::MIME_Filter>, L<MIME::Parser>

=cut
